/****************************************************************************
* Mtrace.java
*
* PURPOSE
* Module that controls the execing and results of the mtrace binary.
*
* MHEALTH
* Written by David Makofske and Kevin Almeroth
* University of California at Santa Barbara
* Department of Computer Science
* Neworking and Multimedia Systems Laboratory
* http://imj.ucsb.edu/mcontrol
*
* COPYRIGHT
* Copyright 1998, 1999
* University of California at Santa Barbara
* Networking and Multimedia Systems Lab
* Santa Barbara, CA  93106
* ALL RIGHTS RESERVED
*
* This material may be modified, copied and redistributed, both within
* the recipient's organization and externally, subject to the following
* restrictions:  (a)  The recipient may not derive income from the
* University of California at Santa Barbara (herein "UCSB") information
* itself;  (b)  In any material based on this information, the recipient
* agrees to acknowledge UCSB; (c)  Any copies made of this material must be
* accompanied by the following copyright notice: "Copyright 1998, 1999
* University of California at Santa Barbara. Santa Barbara, CA 93106. All
* Rights Reserved."; and (d)  The recipient agrees to obey all U.S.
* Government restrictions governing redistribution or export of such
* information. These restrictions may apply to redistribution within an
* international organization. UCSB makes no warranties or representations,
* either expressed or implied, with respect to the results containied
* herein, its quality, merchantability, performance or fitness for a
* particular purpose. In no event shall UCSB or its developers, directors,
* officers, employees or affiliates be liable for direct, incidental,
* indirect, special or consequential damages (including damages or loss
* of business profits, business interruption, loss of business information
* and the like) resulting from any defect in this material or its
* documentation or arising out of the use or inability to use this material
* or accompanying documentation even if UCSB, an authorized representative
* or a UCSB affiliate has been advised of the possibility of such damage.
* UCSB makes no representation or warranty regarding the results obtainable
* through use of this material.  No oral or written information or advice
* given by UCSB, its dealers, distributors, agents, affiliates, developers,
* directors, officers or employees shall create a warranty or in any way
* increase the scope of this warranty.
****************************************************************************/

import java.lang.*; 
import java.io.*; 
import java.util.*; 

public class Mtrace extends Thread {
 
  private SessionData sessionData;

  private Process mtrace_process;
  private MHealth parent;
  private String SourceIP, ReceiverName, ReceiverIP;

  private ReceiverPath receiverPath; 
  private static final int MAX_HOPS = 40;

  private boolean Continue=true;

  private ReceiverNode currMtraceReceiverNode;

  private static final int MTRACE_NORMAL  = 1;
  private static final int MTRACE_DASH_G  = 2;
  private static final int MTRACE_REVERSE = 3;

  public Mtrace(MHealth parent, 
                SessionData sessionData,
                ReceiverNode currMtraceReceiverNode) {

    this.parent = parent;
    this.sessionData = sessionData;
    this.currMtraceReceiverNode = currMtraceReceiverNode;

    this.SourceIP = new String(sessionData.sourceList.getSourceRoot().getIP());
    this.ReceiverName = new String(currMtraceReceiverNode.getName());
    this.ReceiverIP = new String(currMtraceReceiverNode.getIP());

    receiverPath = new ReceiverPath(ReceiverName, ReceiverIP);
  }

  public void run() {
    Continue = true;

    receiverPath.mtraceStart = new Date();

    if (! doNormalMtrace())
      if (! doDashGMtrace())
        doReverseMtrace();

    receiverPath.mtraceFinish = new Date();

    receiverPath.indexLastHop--;
    parent.mtraceDone(receiverPath);
  }

  private boolean doNormalMtrace() {
    BufferedReader br;

    br = doMtrace(MTRACE_NORMAL, null);
    if (br != null) {
      if (parseMtrace(MTRACE_NORMAL, br))
        return(true);
    }
    return(false);
  }

  private boolean doDashGMtrace() {
    BufferedReader br;
    String lasthop;

    // check if we already know this receiver's last hop router:
    lasthop = currMtraceReceiverNode.getLastHopIP();
 
    if (lasthop == null) {
      // we don't know it, try to find it:
      lasthop = getLastHopRouter();
      if (lasthop != null) {
        // we got it, save for future use: 
        currMtraceReceiverNode.setLastHopIP(lasthop);
      }
    }

    // if we got the last hop router, do a gateway mtrace:
    if (lasthop != null) {
      br = doMtrace(MTRACE_DASH_G, lasthop);
      if (br != null) {
        if (parseMtrace(MTRACE_DASH_G, br)) 
          return(true);
      }
    }
    return(false);
  }

  private boolean doReverseMtrace() {
    BufferedReader br;

    br = doMtrace(MTRACE_REVERSE, null);
    if (br != null) {
      if (parseMtrace(MTRACE_REVERSE, br)) 
        return(true);
    }
    return(false);
  }

  public BufferedReader doMtrace(int Type, String lasthop) {
    String execString, str;
    String execArray[];
    Runtime rt;
    byte[] b = new byte[4000];
    BufferedReader br;
    int ArraySize;

    try {
      switch (Type) {
      case MTRACE_NORMAL:
        // do a normal mtrace
        ArraySize = 4;
        execArray = new String[ArraySize];
        execArray[0] = new String("mtrace"); 
        execArray[1] = new String(SourceIP);
        execArray[2] = new String(ReceiverIP);
        execArray[3] = new String(sessionData.getSessionIP());
        break;
      case MTRACE_DASH_G:
        // do a third party mtrace (-g option)
        ArraySize = 6;
        execArray = new String[ArraySize];
        execArray[0] = new String("mtrace"); 
        execArray[1] = new String("-g");
        execArray[2] = new String(lasthop);
        execArray[3] = new String(SourceIP);
        execArray[4] = new String(sessionData.getSessionIP());
        execArray[5] = new String(ReceiverIP);
        break;
      case MTRACE_REVERSE:
        // do a reverse mtrace with a -s (short)
        ArraySize = 5;
        execArray = new String[ArraySize];
        execArray[0] = new String("mtrace"); 
        execArray[1] = new String("-s");
        execArray[2] = new String(ReceiverIP);
        execArray[3] = new String(SourceIP);
        execArray[4] = new String(sessionData.getSessionIP());
        break;
      default:
        return(null);
      } 

      // for debugging, print out all exec args:
      log("");
      log("#################################################################");
      log("# TIMESTAMP START = " + new Date().getTime() + "\n");
      String temp = "";
      for (int x=0; x < ArraySize; x++) {
        temp = temp + execArray[x] + " ";
      }
      log(temp);

      rt = Runtime.getRuntime();
      try {
        // exec mtrace
        if (mtrace_process != null) mtrace_process.destroy();
        mtrace_process = rt.exec(execArray);
      } catch (SecurityException se) {
        se.printStackTrace();
      }

      // read output
      br = new BufferedReader(new InputStreamReader
                              (mtrace_process.getInputStream()));

      br.mark(4000);

      String s="";

      while (Continue) {
        try {
          s = br.readLine();
          if (s == null) 
            break;
          else
            log(s);
        } catch (IOException ioe) { }
      }
         
      
      br.reset();

    } catch(IOException exc) {
      System.out.println(exc.getMessage());
      exc.printStackTrace();
      log("# TIMESTAMP FINISH = " + new Date().getTime() + "\n");
      return(null);
    }
    log("# TIMESTAMP FINISH = " + new Date().getTime() + "\n");
    return(br);
  }
 
  public boolean parseMtrace(int Type, BufferedReader br) {

    if (! parseMtraceRoutes(Type, br)) return(false);

    if (Type != MTRACE_REVERSE) 
       parseMtraceStats(br);

    return(true);
  }

  public boolean parseMtraceRoutes(int Type, BufferedReader br) {
    String line;
    int intLineNum=0;
    String strLineNum;
 
    receiverPath.mtraceHop = new MtraceHop[MAX_HOPS];
    receiverPath.indexLastHop=0;

    try {
      line = br.readLine();
      while ((line != null) && (Continue)) {

        if (intLineNum == 0)
          strLineNum = " 0 ";
        else
          strLineNum = "-" + intLineNum + " ";

        if (line.indexOf(strLineNum) != -1) {
          if (parseMtraceLine(strLineNum,line))
            intLineNum++;
        } else if (line.indexOf("Round trip time") != -1) break;
        line = br.readLine();
      }
    } catch (IOException ioe) {
      ioe.printStackTrace();
    }
    if (Type == MTRACE_REVERSE)
      ReverseArray();

/*    System.out.println("source " + SourceIP + " must = " 
                         + receiverPath.mtraceHop[receiverPath.indexLastHop-1].getIP());
    System.out.println("rcvr " + ReceiverIP + " must = " 
                       + receiverPath.mtraceHop[0].getIP());
    for (int y=0; y<(receiverPath.indexLastHop); y++) {
      System.out.println(y + ") " + receiverPath.mtraceHop[y].getIP());
    }*/

    try {
      if ((SourceIP.equals(receiverPath.mtraceHop[receiverPath.indexLastHop-1].getIP())) &&
          (ReceiverIP.equals(receiverPath.mtraceHop[0].getIP())))
        return(true);
      else
        return(false);
    } catch (ArrayIndexOutOfBoundsException oob) {
      // this will happen if we kill -9 the mtrace
      return(false);
    }
  }

  public boolean parseMtraceLine(String strLineNum, String line) {
    int pos;

    pos = line.indexOf(strLineNum);
    if (pos == -1) 
      return(false);

    pos+=3;

    pos = skipSpaces(pos,line); 
    if (pos == -1) return(false);
    
    String name=getName(pos,line);
    if (name == null) return(false);

    String ip = getValInParen(pos, line);
    if (ip == null) return(false);

    if (name.equals("?")) name = name + " (" + ip + ")";

//    System.out.println("hop " + strLineNum + ": name = " + name + ", ip = " 
//                       + ip);

    if (receiverPath.indexLastHop >= MAX_HOPS) {
      // Uhoh, we are over the limit
      System.out.println("This mtrace has exceeded the max hops " + MAX_HOPS);
      return(false);
    }
    receiverPath.mtraceHop[receiverPath.indexLastHop] = new MtraceHop(name,ip);

    // parse protocol:
    if ((pos = line.indexOf("PIM")) != -1) {
//       System.out.println("PIM type is " + getNextVal(pos,line)); 
       receiverPath.mtraceHop[receiverPath.indexLastHop].setProtocol(getNextVal(pos,line)); 
    } else if (line.indexOf("DVMRP") != -1) {
       receiverPath.mtraceHop[receiverPath.indexLastHop].setProtocol("DVMRP");
    } else if (line.indexOf("Unknown protocol code 8") != -1) {
       receiverPath.mtraceHop[receiverPath.indexLastHop].setProtocol("Unknown Protocol Code 8");
    }

    receiverPath.indexLastHop++;

    return(true);
  }

  public int skipSpaces(int pos, String line) {
    try {
      while ((line.substring(pos,pos+1).equals(" ")) && (Continue)) pos++;
    } catch (StringIndexOutOfBoundsException exc) {
      return(-1);
    }
    return(pos);
  }

  public String getName(int pos, String line) {
    String name="";

    try {
      while ((! line.substring(pos, pos+1).equals(" ")) && (Continue)) {
        name = name + line.substring(pos, pos+1);
        pos++;
      }
      if (name.equals("*")) {
        pos++;
        name = getName(pos, line);
      }
    } catch (StringIndexOutOfBoundsException exc) {
      return(null);
    }
    return(name);
  }

  public String getNextVal(int pos, String line) {
    String val="";

    try {
      while ((! line.substring(pos, pos+1).equals(" ")) && (Continue)) {
        val = val + line.substring(pos, pos+1);
        pos++;
      }
    } catch (StringIndexOutOfBoundsException exc) {
      return(null);
    }
    return(val);
  }

  private String getValInParen(int pos, String line) {
    int pos1, pos2;

    line = line.substring(pos);
    try {
      pos1 = line.indexOf("(");
      if (pos1 == -1) return(null);
      pos2 = line.indexOf(")");
      if (pos2 == -1) return(null);

      return(line.substring(pos1+1,pos2));

    } catch (StringIndexOutOfBoundsException exc) {
      return(null);
    }
  }

  private boolean parseMtraceStats(BufferedReader br) {
    String line;
    int hopnum=receiverPath.indexLastHop-1;
    int pos;
    boolean foundIP=false;

    try {
      line = br.readLine();
      while ((line != null) && (hopnum >= 0) && (Continue)) {

        if (! foundIP) {
          if (line.indexOf(receiverPath.mtraceHop[hopnum].getIP()) != -1)
            foundIP = true;
        } else {
          String loss;
          pos = line.indexOf("/");
          if (pos == -1) 
            loss = "?/?";
          else {
            pos = skipSpaces(pos-4, line);
            if (pos == -1) 
              loss = "?/?";
            else
              loss = getNextVal(pos, line);
          }
          if (loss == null) loss = "?/?";

          if (loss.equals("")) {
            receiverPath.mtraceHop[hopnum].setLostPackets("");
            receiverPath.mtraceHop[hopnum].setTotalPackets("");
          } else {
            receiverPath.mtraceHop[hopnum].setLostPackets(loss.substring(0, loss.indexOf("/")));
            receiverPath.mtraceHop[hopnum].setTotalPackets(loss.substring(loss.indexOf("/")+1));
          }

//          System.out.println(receiverPath.mtraceHop[hopnum].getName() + "/" + 
//                             receiverPath.mtraceHop[hopnum].getIP + " = " + loss);
          
          foundIP = false;
          hopnum--;
        }
        line = br.readLine();
      }
    } catch (IOException ioe) {
      ioe.printStackTrace();
    }
    return(true);
  }

  public String getLastHopRouter() {
    String execString, str;
    String execArray[];
    Runtime rt;
    byte[] b = new byte[2000];
    int ArraySize = 2;
    BufferedReader br;

    try {
      execArray = new String[ArraySize];
      execArray[0] = new String("mtrace"); 
      execArray[1] = new String(ReceiverIP);

      // for debugging, print out all exec args:
      log("");
      log("# TIMESTAMP START = " + new Date().getTime() + "\n");
      String temp = "";
      for (int x=0; x < ArraySize; x++) {
        temp = temp + execArray[x] + " ";
      }
      log(temp);

      rt = Runtime.getRuntime();
      try {
        // exec mtrace
        if (mtrace_process != null) mtrace_process.destroy();
        mtrace_process = rt.exec(execArray);
      } catch (SecurityException se) {
        se.printStackTrace();
      }

      // read output
      br = new BufferedReader(new InputStreamReader
                              (mtrace_process.getInputStream()));

//      br.mark(10000);

      String s="";
      String holdIP="";
      String newIP="";

      while (Continue) {
        try {
          s = br.readLine();
          if (s == null) 
            break;
          else {
            log(s);
            holdIP = newIP;
            newIP = getValInParen(0, s);
            if (ReceiverIP.equals(newIP)) {
              mtrace_process.destroy();
              log("# TIMESTAMP FINISH = " + new Date().getTime() + "\n");
              return(holdIP);
            }
          }
        } catch (IOException ioe) { }
      }
         
    } catch(IOException exc) {
      System.out.println(exc.getMessage());
      exc.printStackTrace();
      log("# TIMESTAMP FINISH = " + new Date().getTime() + "\n");
      return(null);
    }
    log("# TIMESTAMP FINISH = " + new Date().getTime() + "\n");
    return(null);
  }
 
  private void ReverseArray() {
    int end = receiverPath.indexLastHop-1;
    int beg = 0;
    MtraceHop holdhop;
   
    while (end > beg) {
//      System.out.println("Swapping " + beg + " with " + end + ".");
      holdhop = receiverPath.mtraceHop[end];

      receiverPath.mtraceHop[end] = receiverPath.mtraceHop[beg]; 

      receiverPath.mtraceHop[beg] = holdhop;

      end--;
      beg++;
    }
  }

  public void killMtrace() {
    Continue = false;
    if (mtrace_process != null)
      mtrace_process.destroy();
  }

  private void log(String str) {
    // comment out this print to eliminate mtrace to stdout:
    System.out.println(str);
    if (sessionData.getRouteLoggingEnabled()) {
      if (sessionData.RouteLog == null)
        startLogging();  
      try {
        sessionData.RouteLog.write(str + "\n");
        sessionData.RouteLog.flush();
      } catch (IOException ioe) {
        ioe.printStackTrace();
      }
    }
  }

  public void logFileNameDirChanged() {
    // either log filename or directory or both have changed!
    stopLogging();
    startLogging();
  }

  public void startLogging() {
    try {
      // open log file
      sessionData.RouteLog = new FileWriter(sessionData.getLogDir() + "/" +
                                            sessionData.getLogFileName() + 
                                            ".ROUTE.LOG",true);
    } catch (IOException ioe) {
      ioe.printStackTrace();
      return;
    }
    try {
      // output log file headers
      sessionData.RouteLog.write("# ROUTE LOG\n");
      Date date = new Date();
      long millis = date.getTime();
      sessionData.RouteLog.write("# LOG SESSION: " + sessionData.getSessionIP()
                                 + "/" + sessionData.getSessionPort() + "\n");
      sessionData.RouteLog.write("# LOG START TIME: " + date.toString() + 
                                 " = " + millis + "\n");
      sessionData.RouteLog.flush();
    } catch (IOException ioe) {
      ioe.printStackTrace();
    }
  }

  public void stopLogging() {
    if (sessionData.RouteLog != null) {
      try {
        // output log file trailers
        sessionData.RouteLog.write("# ROUTE LOG\n");
        Date date = new Date();
        long millis = date.getTime();
        sessionData.RouteLog.write("# LOG SESSION: " + 
                                   sessionData.getSessionIP() + "/" +
                                   sessionData.getSessionPort() + "\n");
        sessionData.RouteLog.write("# LOG STOP TIME: " + date.toString() + 
                                   " = " + millis + "\n");
        sessionData.RouteLog.write("#####################################################\n");
        sessionData.RouteLog.flush();
        sessionData.RouteLog.close();
      } catch (IOException ioe) {
        ioe.printStackTrace();
      }
      sessionData.RouteLog = null;
    }
  }
}
