/****************************************************************************
* ReceiverList.java
*
* PURPOSE
* Class to hold listen of multicast receivers.
*
* MHEALTH
* Written by David Makofske and Kevin Almeroth
* University of California at Santa Barbara
* Department of Computer Science
* Neworking and Multimedia Systems Laboratory
* http://imj.ucsb.edu/mcontrol
*
* COPYRIGHT
* Copyright 1998, 1999
* University of California at Santa Barbara
* Networking and Multimedia Systems Lab
* Santa Barbara, CA  93106
* ALL RIGHTS RESERVED
*
* This material may be modified, copied and redistributed, both within
* the recipient's organization and externally, subject to the following
* restrictions:  (a)  The recipient may not derive income from the
* University of California at Santa Barbara (herein "UCSB") information
* itself;  (b)  In any material based on this information, the recipient
* agrees to acknowledge UCSB; (c)  Any copies made of this material must be
* accompanied by the following copyright notice: "Copyright 1998, 1999
* University of California at Santa Barbara. Santa Barbara, CA 93106. All
* Rights Reserved."; and (d)  The recipient agrees to obey all U.S.
* Government restrictions governing redistribution or export of such
* information. These restrictions may apply to redistribution within an
* international organization. UCSB makes no warranties or representations,
* either expressed or implied, with respect to the results containied
* herein, its quality, merchantability, performance or fitness for a
* particular purpose. In no event shall UCSB or its developers, directors,
* officers, employees or affiliates be liable for direct, incidental,
* indirect, special or consequential damages (including damages or loss
* of business profits, business interruption, loss of business information
* and the like) resulting from any defect in this material or its
* documentation or arising out of the use or inability to use this material
* or accompanying documentation even if UCSB, an authorized representative
* or a UCSB affiliate has been advised of the possibility of such damage.
* UCSB makes no representation or warranty regarding the results obtainable
* through use of this material.  No oral or written information or advice
* given by UCSB, its dealers, distributors, agents, affiliates, developers,
* directors, officers or employees shall create a warranty or in any way
* increase the scope of this warranty.
****************************************************************************/

import java.util.*;
import java.io.*;

public class ReceiverList {

  private ReceiverNode root=null;

  private ReceiverNode currentMtrace = null;
  private ReceiverNode nextMtrace = null;

  private RtcpLogger rtcplogger = null;

  public ReceiverList() {
    root = null;
  }

  public void addReceiverNode(RtcpReport report) {
    if (root == null) {
      ReceiverNode rn = new ReceiverNode(report);
      root=rn;
    } else {
      ReceiverNode rn = new ReceiverNode(report);
      ReceiverNode tail = getReceiverTail(getReceiverRoot());

      if (tail == null) {
        // List is empty, make this the new root
        root = new ReceiverNode(report);
      } else {
        // Append to tail of list
        ReceiverNode newNode = new ReceiverNode(report);
        newNode.setPrevNode(tail);
        tail.setNextNode(newNode);
      }
    }
  }

  public void removeReceiverNode(RtcpReport report) {
    ReceiverNode rn;

    rn = findReceiverNode(getReceiverRoot(), report.getIP());

    if (rn == null) return;

    checkMtracePointers(rn); // move any pointer pointed at removing node

    if (rn == getReceiverRoot()) {
      // remove root of list
      if (rn.getNextNode() == null) {
        // root is only member
        root=null;
      } else {
        rn.getNextNode().setPrevNode(null);
        root = rn.getNextNode();
      } 
    } else {
      if (rn.getNextNode() == null) {
        rn.getPrevNode().setNextNode(null);
      } else {
        rn.getPrevNode().setNextNode(rn.getNextNode());
        rn.getNextNode().setPrevNode(rn.getPrevNode());
      }
    }
  }

  public ReceiverNode getReceiverRoot() { return root; }


  public boolean isReceiverRoot(ReceiverNode rn) { 
    if (root == null) 
      return false;

    return(rn.getIP().equals(root.getIP()));
  }

  public void timeoutReceiverInList(String IP) {
    ReceiverNode rn;

    rn = findReceiverNode(getReceiverRoot(), IP);

    if (rn != null) {
      rn.setTimeout(true);
    }
  }


  public ReceiverNode findReceiverNode(ReceiverNode start, String IP) {
    ReceiverNode next;

    if (start == null) 
      return(null); 
    else if (IP.equals(start.getIP())) 
      return(start);
    else
      return(findReceiverNode(start.getNextNode(), IP));
  }

  public ReceiverNode getReceiverTail(ReceiverNode start) { 
    if (start == null) 
      return(null);
    else if (start.getNextNode() == null)
      return(start);
    else
      return(getReceiverTail(start.getNextNode()));
  }

  public ReceiverNode getReceiverToMtrace() {
    ReceiverNode ret=null;

    if (nextMtrace != null) {
      // this is a user requested mtrace
      ret = nextMtrace;
      nextMtrace = null;
    } else {
      if (currentMtrace != null) {
        currentMtrace = currentMtrace.getNextNode();
        ret = currentMtrace;
      } else {
        currentMtrace = root;
        ret = root;
      }
    } 
    return(ret);
  }

  public void setNextMtrace(String IP) {
    ReceiverNode hold;

    hold = findReceiverNode(getReceiverRoot(), IP);
    if (hold != null) {
      if (nextMtrace != null) {
        // cancel another user-requested mtrace
        setMessage(nextMtrace.getIP(),"Mtrace: CANCELLED");
      }
      setMessage(hold.getIP(),"Mtrace: SCHEDULED");
      nextMtrace = hold; 
    }
  }

  private void checkMtracePointers(ReceiverNode rn) {
    if (rn.equals(nextMtrace)) {
      // this was a user scheduled mtrace, cancel it
      nextMtrace = null;
    }
    if (rn.equals(currentMtrace)) {
      // this is the standard trace loop, jump to next node
      currentMtrace = currentMtrace.getNextNode();
    }
  }

  public boolean setMessage(String IP, String msg) {
    ReceiverNode rn = findReceiverNode(getReceiverRoot(), IP);
    if (rn != null) {
      rn.setMessage(msg);
      return(true);
    } else
      return(false);
  }

  public ReceiverNode findReceiverInside(ReceiverNode start, 
                                         int mouseX, int mouseY) {
    
    if (start == null) return(null);

    if (start.isInside(mouseX, mouseY)) return(start); 

    return(findReceiverInside(start.getNextNode(), mouseX, mouseY));
  }

  public boolean isEmpty() {
    return(root == null);
  }

  public void setRtcpLogger(RtcpLogger rtcplogger) {
    this.rtcplogger = rtcplogger;
  }

  public RtcpLogger getRtcpLogger() {
    return(rtcplogger);
  }

  public void delete() {
    while (root != null) {
      removeReceiverNode(root.getRtcpReport());
    }
  }

}
