/****************************************************************************
* RtcpLogger.java
*
* PURPOSE
* Log received RTCP data packets.
*
* MHEALTH
* Written by David Makofske and Kevin Almeroth
* University of California at Santa Barbara
* Department of Computer Science
* Neworking and Multimedia Systems Laboratory
* http://imj.ucsb.edu/mcontrol
*
* COPYRIGHT
* Copyright 1998, 1999
* University of California at Santa Barbara
* Networking and Multimedia Systems Lab
* Santa Barbara, CA  93106
* ALL RIGHTS RESERVED
*
* This material may be modified, copied and redistributed, both within
* the recipient's organization and externally, subject to the following
* restrictions:  (a)  The recipient may not derive income from the
* University of California at Santa Barbara (herein "UCSB") information
* itself;  (b)  In any material based on this information, the recipient
* agrees to acknowledge UCSB; (c)  Any copies made of this material must be
* accompanied by the following copyright notice: "Copyright 1998, 1999
* University of California at Santa Barbara. Santa Barbara, CA 93106. All
* Rights Reserved."; and (d)  The recipient agrees to obey all U.S.
* Government restrictions governing redistribution or export of such
* information. These restrictions may apply to redistribution within an
* international organization. UCSB makes no warranties or representations,
* either expressed or implied, with respect to the results containied
* herein, its quality, merchantability, performance or fitness for a
* particular purpose. In no event shall UCSB or its developers, directors,
* officers, employees or affiliates be liable for direct, incidental,
* indirect, special or consequential damages (including damages or loss
* of business profits, business interruption, loss of business information
* and the like) resulting from any defect in this material or its
* documentation or arising out of the use or inability to use this material
* or accompanying documentation even if UCSB, an authorized representative
* or a UCSB affiliate has been advised of the possibility of such damage.
* UCSB makes no representation or warranty regarding the results obtainable
* through use of this material.  No oral or written information or advice
* given by UCSB, its dealers, distributors, agents, affiliates, developers,
* directors, officers or employees shall create a warranty or in any way
* increase the scope of this warranty.
****************************************************************************/

import java.net.*;
import java.lang.*;
import java.util.*;
import java.io.*;

public class RtcpLogger extends Thread implements TimerListener {

  private SessionData sessionData;

//  private static final int TURN_GRAY_TIME = 30000;  // in milliseconds
//  private static final int TIMEOUT_TIME   = 60000;  // in milliseconds
  private static final int TURN_GRAY_TIME = 120000;  // in milliseconds
  private static final int TIMEOUT_TIME   = 240000;  // in milliseconds

  private boolean RUNNING;

  private byte buf[];
  private int byte_offset;
  private FileWriter fw = null;

  private boolean byePacketReceived = false;

  private RtcpReportHeader header;

  private RtcpReport holdReport;
  private Hashtable sender_reports=null;
  private Hashtable receiver_reports=null;
  private String senderIP;
  private long senderSsrc;

  private int packetType;
  private int packetLength;

  private int existingStatus = -2;

  private Timer timer;

  private Vector listener=null;

  private final static int EVENT_SENDER_ADDED            = 0;
  private final static int EVENT_SENDER_REMOVED          = 1;
  private final static int EVENT_SENDER_UPDATED          = 2;
  private final static int EVENT_SENDER_UPDATED_REDRAW   = 3;
  private final static int EVENT_RECEIVER_ADDED          = 4;
  private final static int EVENT_RECEIVER_REMOVED        = 5;
  private final static int EVENT_RECEIVER_UPDATED        = 6;
  private final static int EVENT_RECEIVER_UPDATED_REDRAW = 7;
  private final static int EVENT_RECEIVER_TIMEOUT        = 8;

  public RtcpLogger(SessionData sessionData) {
    //System.out.println("new RtcpLogger(" + sessionData.getSessionIP() 
    //                   + "," + sessionData.getSessionPort() + "," + 
    //                   sessionData.getRtcpLoggingEnabled() + ")");

    this.sessionData = sessionData;

    sender_reports = new Hashtable();
    receiver_reports = new Hashtable();

    listener = new Vector();
  }

  public void run() {
   
    RUNNING=true;

    MulticastSocket s;
    DatagramPacket dp;
    int i;
    boolean parseSuccessful;

    timer = new Timer(10000); // 10 seconds
    timer.addTimerListener(this);

    try {
      s = new MulticastSocket(Integer.parseInt(sessionData.getSessionPort())+1);
      //Debug.out(debugLevel, Debug.INFO, "RtcpLogger: new MulticastSocket on " +
      //          "port: " + s.getLocalPort());

      // join the multicast group
      s.joinGroup(InetAddress.getByName(sessionData.getSessionIP()));
    } catch (Exception e) {
      e.printStackTrace();
      return;
    }

    while(RUNNING) {
      try {
        buf = new byte[1000];
        dp = new DatagramPacket(buf, buf.length);

        //Debug.out(debugLevel, Debug.INFO, "RtcpLogger: blocking for packet");

        s.receive(dp);

        //Debug.out(debugLevel, Debug.INFO, "RtcpLogger: packet received from " +
        //          dp.getAddress().toString() + "/" + dp.getPort() + ", len = "
        //          + dp.getLength());

        Integer port = new Integer(dp.getPort());
        String addr = new String(dp.getAddress().toString());
        int loc = addr.indexOf("/");
        if (loc == -1) {
          parseSuccessful = parsePacket(addr,addr,port.toString(),dp.getLength());
        } else {
          String name = new String(addr.substring(0,loc));
          addr = new String(addr.substring(loc+1));
          parseSuccessful = parsePacket(name,addr,port.toString(),dp.getLength());
        }
        if (parseSuccessful) {
          savePacket();
          if (sessionData.getRtcpLoggingEnabled()) logPacket();
        }
      } catch (Exception e) {
        e.printStackTrace();
      }
    }
/*    try {
      s.leaveGroup(InetAddress.getByName(sessionData.getSessionIP()));
      s.close();
    } catch (Exception e) {
      e.printStackTrace();
    }*/
  }
 
  public boolean parsePacket(String name, String ip, String port, int tot_len) {

    holdReport = findExistingReport(ip);
    if (holdReport == null) 
      holdReport = new RtcpReport(name, ip, port);

    byte_offset = 0;

    while (byte_offset < tot_len) {

      if (! parseHeader()) 
        return(false);

      switch(header.getPacketType()) {
      case -1:
        // unable to determine packet type
        return(false);
      case 200:
        //System.out.println("200");
        if (! parseSenderPacket200()) 
          return(false);
        break;
      case 201:
        //System.out.println("201");
        if (! parseReceiverPacket201())
          return(false);
        break;
      case 202:
        //System.out.println("202");
        if (! parseSdesPacket202()) 
          return(false);
        break;
      case 203:
        //System.out.println("203");
        parseByePacket203();
        return(false); // return false so we don't bother saving
      case 204:
        //System.out.println("204");
        if (! parseAppPacket204())
          return(false);
        break;
      default:
        return(false);
      }
    }
    return(true);
  }

  private boolean parseSenderPacket200() {
    holdReport.setHeader(header);

    parseSenderInfo();

    if (! parseReportBlocks())
      return(false);

    return(true);
  }

  private boolean parseReceiverPacket201() {
    holdReport.setHeader(header);

    if (! parseReportBlocks())
      return(false);
  
    return(true);
  }

  private boolean parseSdesPacket202() {
//    parseSdesHeader();
    parseSdesPacket();

    //Debug.out(debugLevel, Debug.DATA, "RtcpLogger: " + 
    //          "***********************************************************");

    return(true);
  }

  private boolean parseByePacket203() {
    byePacketReceived = true;
    receiver_reports.remove(holdReport.getIP());
    eventNotify(holdReport, EVENT_RECEIVER_REMOVED);
      
    return(true);
  }

  private boolean parseAppPacket204() {
    return(true);
  }

  public boolean parseHeader() {
    byte hold;
    byte hold_array[];
    byte displayMask;
    Byte b;

    hold = buf[byte_offset];
    byte_offset++;
    displayMask = (byte)(1 << 7);
    String version = new String((hold & displayMask) == 0 ? "0" : "1");
    displayMask = (byte)(1 << 6);
    version = version + ((hold & displayMask) == 0 ? "0" : "1");
    //Debug.out(debugLevel, Debug.DATA, "RtcpLogger: version=" + version);
    if (! version.equals("10")) {
      //Debug.out(debugLevel, Debug.ERROR, "RtcpLogger: version " + 
      //          version + " invalid");
      return(false);
    }

    displayMask = (byte)(1 << 5);
    String pad = new String((hold & displayMask) == 0 ? "0" : "1");
    //System.out.println("RtcpLogger: pad=" + pad);

    displayMask = (byte)31;
    b = new Byte((byte)(hold & displayMask));
    int count_or_subtype = b.intValue();
    //System.out.println("RtcpLogger: count or subtype=" + 
    //                              count_or_subtype);

    ParseData parseData = new ParseData();

    hold_array = new byte[1];
    hold_array[0] = buf[byte_offset];
    byte_offset++;
    int packet_type = parseData.intValue(1,hold_array);
    //Debug.out(debugLevel, Debug.DATA, "RtcpLogger: packet type=" + 
    //          packet_type);
    if (! isValidPacketType(packet_type)) {
      //Debug.out(debugLevel, Debug.ERROR, "RtcpLogger: invalid packet type " + 
      //          packet_type);
      return(false);
    }

    hold_array = new byte[2];
    for (int i=0; i<2; i++) hold_array[i] = buf[i+byte_offset];
    byte_offset += 2;
    int packet_len = parseData.intValue(2,hold_array);
    packetLength = 4 * packet_len;
    //Debug.out(debugLevel, Debug.DATA, "RtcpLogger: packet length=" + 
    //          packet_len);

    hold_array = new byte[4];
    for (int i=0; i<4; i++) hold_array[i] = buf[i+byte_offset];
    byte_offset += 4;
    long ssrc = parseData.longValue(4,hold_array);
    //Debug.out(debugLevel, Debug.DATA, "RtcpLogger: ssrc=" + ssrc);

    header = new RtcpReportHeader(version, pad, count_or_subtype, packet_type,
                                  packet_len, ssrc);
    return(true);
  }

  public void parseSenderInfo() {
    byte hold_array[];
    int i;

    //Debug.out(debugLevel, Debug.DATA, "RtcpLogger: *** Sender Info ***");

    ParseData parseData = new ParseData();

    hold_array = new byte[4];
    for (i=0; i<4; i++) hold_array[i] = buf[i+byte_offset];
    byte_offset+=4;
    long ntp_timestamp_most = parseData.longValue(4,hold_array);
    //Debug.out(debugLevel, Debug.DATA, "RtcpLogger: ntp timestamp most=" + 
    //          ntp_timestamp_most);

    hold_array = new byte[4];
    for (i=0; i<4; i++) hold_array[i] = buf[i+byte_offset];
    byte_offset+=4;
    long ntp_timestamp_least = parseData.longValue(4,hold_array);
    //Debug.out(debugLevel, Debug.DATA, "RtcpLogger: ntp timestamp least=" + 
    //          ntp_timestamp_least);

    hold_array = new byte[4];
    for (i=0; i<4; i++) hold_array[i] = buf[i+byte_offset];
    byte_offset+=4;
    long rtp_timestamp = parseData.longValue(4,hold_array);
    //Debug.out(debugLevel, Debug.DATA, "RtcpLogger: rtp timestamp=" + 
    //          rtp_timestamp);

    hold_array = new byte[4];
    for (i=0; i<4; i++) hold_array[i] = buf[i+byte_offset];
    byte_offset+=4;
    long sender_packet_count = parseData.longValue(4,hold_array);
    //Debug.out(debugLevel, Debug.DATA, "RtcpLogger: sender packet count=" + 
    //          sender_packet_count);

    hold_array = new byte[4];
    for (i=0; i<4; i++) hold_array[i] = buf[i+byte_offset];
    byte_offset+=4;
    long sender_octet_count = parseData.longValue(4,hold_array);
    //Debug.out(debugLevel, Debug.DATA, "RtcpLogger: sender octet count=" + 
    //          sender_octet_count);

    holdReport.setSendInfo(ntp_timestamp_most, ntp_timestamp_least,
                            rtp_timestamp, sender_packet_count,
                            sender_octet_count);

    return;
  }

  public boolean parseReportBlocks() {
    byte hold_array[];
    int i;

    ParseData parseData = new ParseData();

    for (int count=0; count < holdReport.getReportCount(); count++) {

      //Debug.out(debugLevel, Debug.DATA, "RtcpLogger: *** Report Block " + 
      //          (count+1));

      hold_array = new byte[4];
      for (i=0; i<4; i++) hold_array[i] = buf[i+byte_offset];
      byte_offset+=4;
      holdReport.report_blocks[count].source_ssrc = 
                                          parseData.longValue(4,hold_array);
      //Debug.out(debugLevel, Debug.DATA, "RtcpLogger: source ssrc=" + 
      //          holdReport.report_blocks[count].source_ssrc);

      hold_array = new byte[1];
      hold_array[0] = buf[byte_offset];
      byte_offset++;
      holdReport.report_blocks[count].fraction_lost = 
                                           parseData.intValue(1,hold_array);
//     System.out.println("RtcpLogger: fraction lost u=" + 
//                holdReport.report_blocks[count].fraction_lost);
     
/*     holdReport.report_blocks[count].fraction_lost =  (int)hold_array[0];
     System.out.println("RtcpLogger: fraction lost s=" + 
                holdReport.report_blocks[count].fraction_lost);*/

      hold_array = new byte[3];
      for (i=0; i<3; i++) hold_array[i] = buf[i+byte_offset];
      byte_offset+=3;
      holdReport.report_blocks[count].cum_packets_lost = 
                                    parseData.intValue(3,hold_array);
      //Debug.out(debugLevel, Debug.DATA, "RtcpLogger: cum. packets lost=" + 
      //          holdReport.report_blocks[count].cum_packets_lost);

      hold_array = new byte[2];
      for (i=0; i<2; i++) hold_array[i] = buf[i+byte_offset];
      byte_offset+=2;
      holdReport.report_blocks[count].seq_no_cycles = 
                                     parseData.intValue(2,hold_array);
      //Debug.out(debugLevel, Debug.DATA, "RtcpLogger: seq no cycles=" + 
      //          holdReport.report_blocks[count].seq_no_cycles);

      hold_array = new byte[2];
      for (i=0; i<2; i++) hold_array[i] = buf[i+byte_offset];
      byte_offset+=2;
      holdReport.report_blocks[count].highest_seq_no = 
                                    parseData.intValue(2,hold_array);
      //Debug.out(debugLevel, Debug.DATA, "RtcpLogger: highest seq no=" + 
      //          holdReport.report_blocks[count].highest_seq_no);

      hold_array = new byte[4];
      for (i=0; i<4; i++) hold_array[i] = buf[i+byte_offset];
      byte_offset+=4;
      holdReport.report_blocks[count].interarrival_jitter = 
                              parseData.longValue(4,hold_array);
      //Debug.out(debugLevel, Debug.DATA, "RtcpLogger: interarrival jitter=" + 
      //          holdReport.report_blocks[count].interarrival_jitter);

      hold_array = new byte[4];
      for (i=0; i<4; i++) hold_array[i] = buf[i+byte_offset];
      byte_offset+=4;
      holdReport.report_blocks[count].last_sr =  
                              parseData.longValue(4,hold_array);
      //Debug.out(debugLevel, Debug.DATA, "RtcpLogger: last sr=" + 
      //          holdReport.report_blocks[count].last_sr);

      hold_array = new byte[4];
      for (i=0; i<4; i++) hold_array[i] = buf[i+byte_offset];
      byte_offset+=4;
      holdReport.report_blocks[count].delay_since_lsr = 
                            parseData.longValue(4,hold_array);
      //Debug.out(debugLevel, Debug.DATA, "RtcpLogger: delay since last sr=" + 
      //          holdReport.report_blocks[count].delay_since_lsr);
    }
    return(true);
  }

  public void parseSdesPacket() {
    byte hold_array[];
    int i, adjusted_offset;

    //Debug.out(debugLevel, Debug.DATA, "RtcpLogger: *** SDES Section ***");

    adjusted_offset = byte_offset - 4;
    while ((byte_offset - adjusted_offset) < packetLength) {
      Byte b = new Byte(buf[byte_offset++]);
      int fieldId = b.intValue();

      b = new Byte(buf[byte_offset++]);
      int length = b.intValue();

      if (length < 0) {
        System.out.println("Could not parse SDES packet from " + 
                   holdReport.getIP() + ", invalid length of " + length);
        break;
      }
      hold_array = new byte[length];
      for (i=0; i<length; i++) hold_array[i] = buf[i+byte_offset];
      byte_offset+=length;
      String value = new String(hold_array);

      //System.out.println("RtcpLogger: field(" + fieldId + 
      //          ") of length " + length + " has value of <" + value + ">");

      if (fieldId != 0) 
        holdReport.setSdesInfo(fieldId, value);
    }
  }

  public void savePacket() {
    RtcpReport temp;

    switch (holdReport.getPacketType()) {
    case 200:
      //System.out.println("save 200 packet");
      if (existingStatus == -2) {
        // new report
        sender_reports.put(holdReport.getIP(),holdReport);
        eventNotify(holdReport, EVENT_SENDER_ADDED);
      } else {
        holdReport.setTimeout(false);

        if (holdReport.getStatus(senderSsrc) != existingStatus) 
          eventNotify(holdReport, EVENT_SENDER_UPDATED);
        else
          eventNotify(holdReport, EVENT_SENDER_UPDATED);
      }
      break;
    case 201:
      //System.out.println("save 201 packet");

      // bail out if also a sender
      temp = (RtcpReport)sender_reports.get(holdReport.getIP());
      if (temp != null) return;
      
      if (existingStatus == -2) {
        // new report
        receiver_reports.put(holdReport.getIP(),holdReport);
        eventNotify(holdReport, EVENT_RECEIVER_ADDED);
      } else {
        holdReport.setTimeout(false);

        if (holdReport.getStatus(senderSsrc) != existingStatus)
          eventNotify(holdReport, EVENT_RECEIVER_UPDATED_REDRAW);
        else
          eventNotify(holdReport, EVENT_RECEIVER_UPDATED);
      }
      break;
    }
  }

  public void logPacket() {
    String str;
    Date date;
    long millis;

    if (fw == null) {
      startLogging();
    }

    int rpt_count=0;

    try {
      date = new Date();
      millis = date.getTime();
      str = millis + " " +
            holdReport.getIP() + " " +
            holdReport.getSsrc() + " " +
            holdReport.getReportCount() + " ";
      for (int x=0; x < holdReport.getReportCount(); x++) {
        str = str + holdReport.report_blocks[x].fraction_lost + " " +
            holdReport.report_blocks[x].interarrival_jitter + " ";
      }
      if (byePacketReceived) {
        byePacketReceived = false;
        str = str + "B";
      }
      fw.write(str + "\n");
      fw.flush();
    } catch (IOException ioe) {
      ioe.printStackTrace();
    }
  }

  public Hashtable getSenderReports() { return(sender_reports);
  }

  public Hashtable getReceiverReports() {
    return(receiver_reports);
  }

  public void timeElapsed(TimerEvent evt) {
    RtcpReport hold;
    Date current_timestamp;
    long current_millis;

    // Loop through receivers and check for timeout:

    for (Enumeration e = receiver_reports.elements(); e.hasMoreElements(); ) {

      hold = (RtcpReport)e.nextElement(); 

      current_timestamp = new Date();
      current_millis = current_timestamp.getTime();

      if ((current_millis - hold.getLastUpdate().getTime()) > TIMEOUT_TIME) {
        receiver_reports.remove(hold.getIP());
        eventNotify(hold, EVENT_RECEIVER_REMOVED);
      } else if ((current_millis - hold.getLastUpdate().getTime()) > TURN_GRAY_TIME) {
        hold.setTimeout(true);
        eventNotify(hold, EVENT_RECEIVER_TIMEOUT);
      }
    }
  }

  public void addRtcpLoggerListener(RtcpLoggerListener l) {
     listener.addElement(l);
  }

  public void removeRtcpLoggerListener(RtcpLoggerListener l) {
     listener.removeElement(l);
  }

  private void eventNotify(RtcpReport Report, int Event) {

    //Debug.out(debugLevel, Debug.INFO, "RtcpLogger: eventNotify(" + Event + ")");

    if (listener == null) return; // no one to notify!

    if (listener.size() == 0) return; // no one to notify!

    switch(Event) {
    case EVENT_SENDER_ADDED:
      for (Enumeration e = listener.elements(); e.hasMoreElements(); ) {
        RtcpLoggerListener rll = (RtcpLoggerListener) e.nextElement();
        rll.senderAdded(Report);
      }
      break;
    case EVENT_SENDER_REMOVED:
      for (Enumeration e = listener.elements(); e.hasMoreElements(); ) {
        RtcpLoggerListener rll = (RtcpLoggerListener) e.nextElement();
        rll.senderRemoved(Report);
      }
      break;
    case EVENT_SENDER_UPDATED:
      for (Enumeration e = listener.elements(); e.hasMoreElements(); ) {
        RtcpLoggerListener rll = (RtcpLoggerListener) e.nextElement();
        rll.senderUpdated(Report,false);
      }
      break;
    case EVENT_SENDER_UPDATED_REDRAW:
      for (Enumeration e = listener.elements(); e.hasMoreElements(); ) {
        RtcpLoggerListener rll = (RtcpLoggerListener) e.nextElement();
        rll.senderUpdated(Report,true);
      }
      break;
    case EVENT_RECEIVER_ADDED:
      for (Enumeration e = listener.elements(); e.hasMoreElements(); ) {
        RtcpLoggerListener rll = (RtcpLoggerListener) e.nextElement();
        rll.receiverAdded(Report);
      }
      break;
    case EVENT_RECEIVER_REMOVED:
      for (Enumeration e = listener.elements(); e.hasMoreElements(); ) {
        RtcpLoggerListener rll = (RtcpLoggerListener) e.nextElement();
        rll.receiverRemoved(Report);
      }
      break;
    case EVENT_RECEIVER_UPDATED:
      for (Enumeration e = listener.elements(); e.hasMoreElements(); ) {
        RtcpLoggerListener rll = (RtcpLoggerListener) e.nextElement();
        rll.receiverUpdated(Report, false);
      }
      break;
    case EVENT_RECEIVER_UPDATED_REDRAW:
      // forces a redraw
      for (Enumeration e = listener.elements(); e.hasMoreElements(); ) {
        RtcpLoggerListener rll = (RtcpLoggerListener) e.nextElement();
        rll.receiverUpdated(Report, true);
      }
      break;
    case EVENT_RECEIVER_TIMEOUT:
      for (Enumeration e = listener.elements(); e.hasMoreElements(); ) {
        RtcpLoggerListener rll = (RtcpLoggerListener) e.nextElement();
        rll.receiverTimeout(Report);
      }
      break;
    }
  }


  public boolean isValidPacketType(int packetType) {
    if ((packetType == 200) || (packetType == 201) || 
        (packetType == 202) || (packetType == 203) || 
        (packetType == 204))
      return(true);
    else
      return(false);
  }

  public RtcpReport findExistingReport(String IP) {
    RtcpReport existing;
    existingStatus=-2;

    // check if report is a sender:
    existing = (RtcpReport)sender_reports.get(IP);
    if (existing != null) {
      existingStatus=existing.getStatus(senderSsrc);
      return(existing);
    }
    // check if report is a receiver:
    existing = (RtcpReport)receiver_reports.get(IP);
    if (existing != null) { 
      existingStatus=existing.getStatus(senderSsrc);
      return(existing);
    }
    return(null);
  }

  public void startLogging() {
      try {
        // open log file
        fw = new FileWriter(sessionData.getLogDir() + "/" +
                            sessionData.getLogFileName() + ".RTCP.LOG",true);
      } catch (IOException ioe) {
        ioe.printStackTrace();
        return;
      }
      try {
        // output log file headers
        fw.write("# RTCP PACKET LOG\n");
        Date date = new Date();
        long millis = date.getTime();
        fw.write("# LOG SESSION: " + sessionData.getSessionIP() + "/" +
                 sessionData.getSessionPort() + "\n");
        fw.write("# LOG START TIME: " + date.toString() + " = " +
                 millis + "\n");
        fw.write("# LOG FORMAT: 1-timestamp 2-ip 3-ssrc 4-reportcount" +
                 " n-fractionlost#n n-jitter#n\n");
        fw.flush();
      } catch (IOException ioe) {
        ioe.printStackTrace();
      }
  }

  public void stopLogging() {
    if (fw != null) {
      try {
        // output log file trailers
        fw.write("# RTCP PACKET LOG\n");
        Date date = new Date();
        long millis = date.getTime();
        fw.write("# LOG SESSION: " + sessionData.getSessionIP() + "/" +
                 sessionData.getSessionPort() + "\n");
        fw.write("# LOG STOP TIME: " + date.toString() + " = " +
                 millis + "\n");
        fw.write("##########################################################");
        fw.flush();
        fw.close();
      } catch (IOException ioe) {
        ioe.printStackTrace();
      }
      fw = null;
    }
  }

  public void logFileNameDirChanged() {
    // either log filename or directory or both have changed!
    stopLogging();
    startLogging();
  }

  public void die() {
    RUNNING=false;
  }
}
