/****************************************************************************
* RtcpReport.java
*
* PURPOSE
* Class to hold RTCP report data.
*
* MHEALTH
* Written by David Makofske and Kevin Almeroth
* University of California at Santa Barbara
* Department of Computer Science
* Neworking and Multimedia Systems Laboratory
* http://imj.ucsb.edu/mcontrol
*
* COPYRIGHT
* Copyright 1998, 1999
* University of California at Santa Barbara
* Networking and Multimedia Systems Lab
* Santa Barbara, CA  93106
* ALL RIGHTS RESERVED
*
* This material may be modified, copied and redistributed, both within
* the recipient's organization and externally, subject to the following
* restrictions:  (a)  The recipient may not derive income from the
* University of California at Santa Barbara (herein "UCSB") information
* itself;  (b)  In any material based on this information, the recipient
* agrees to acknowledge UCSB; (c)  Any copies made of this material must be
* accompanied by the following copyright notice: "Copyright 1998, 1999
* University of California at Santa Barbara. Santa Barbara, CA 93106. All
* Rights Reserved."; and (d)  The recipient agrees to obey all U.S.
* Government restrictions governing redistribution or export of such
* information. These restrictions may apply to redistribution within an
* international organization. UCSB makes no warranties or representations,
* either expressed or implied, with respect to the results containied
* herein, its quality, merchantability, performance or fitness for a
* particular purpose. In no event shall UCSB or its developers, directors,
* officers, employees or affiliates be liable for direct, incidental,
* indirect, special or consequential damages (including damages or loss
* of business profits, business interruption, loss of business information
* and the like) resulting from any defect in this material or its
* documentation or arising out of the use or inability to use this material
* or accompanying documentation even if UCSB, an authorized representative
* or a UCSB affiliate has been advised of the possibility of such damage.
* UCSB makes no representation or warranty regarding the results obtainable
* through use of this material.  No oral or written information or advice
* given by UCSB, its dealers, distributors, agents, affiliates, developers,
* directors, officers or employees shall create a warranty or in any way
* increase the scope of this warranty.
****************************************************************************/

import java.io.*;
import java.util.*;

public class RtcpReport {

  // Loss values returned:
  public static final int UNKNOWN     = -1;
  public static final int NO_LOSS     =  0;
  public static final int MEDIUM_LOSS =  1;
  public static final int HIGH_LOSS   =  2;

  // Report Key and General Data:
  private String  name;
  private String  ip;
  private String  port;
  private Date    last_update;
  private boolean timeout = false;

  // RTCP Session Description information
  private String sdesCname;
  private String sdesName;
  private String sdesEmail;
  private String sdesPhone;
  private String sdesLoc;
  private String sdesTool;
  private String sdesNote;

  public RtcpReportHeader   header;
  public RtcpReportSendInfo sender_info;
  public RtcpReportBlock    report_blocks[];

  public RtcpReport(String name, String ip, String port) {
    last_update = new Date();
    this.name   = name;
    this.ip     = ip;
    this.port   = port;
  }

  /************************************************************************/
  // Set and Get Key Fields 
  /************************************************************************/
  public String getName() { return(name); }
  public String getIP() { return(ip); }
  public String getPort() { return(port); }

  /************************************************************************/
  // Set and Get General Fields 
  /************************************************************************/
  public Date getLastUpdate() { return(last_update); }

  public boolean getTimeout() { return(timeout); }
  public void setTimeout(boolean val) { timeout = val; }

  /************************************************************************/
  // Set Header fields
  /************************************************************************/

  public void setHeader(RtcpReportHeader parsedheader) {
    header = new RtcpReportHeader(parsedheader);

    if (header.getReportCount() > 0) {
      report_blocks = new RtcpReportBlock[header.getReportCount()];
      for(int x=0; x < header.getReportCount(); x++) {
        report_blocks[x] = new RtcpReportBlock();
      }
    }
  }

  public String getVersion() { return(header.getVersion()); }
  public String getPad() { return(header.getPad()); }
  public int getPacketType() { return(header.getPacketType());   }
  public int getPacketLen() { return(header.getPacketLen());   }
  public long getSsrc() { return(header.getSsrc());   }
  public int getReportCount() { return(header.getReportCount());   }
  public int getSourceCount() { return(header.getSourceCount());   }
  public int getSubtype() { return(header.getSubtype());   }

  /************************************************************************/
  // Set Sender Info fields
  /************************************************************************/

  public void setSendInfo(long ntp_timestamp_most, long ntp_timestamp_least,
                          long rtp_timestamp, long sender_packet_count, 
                          long sender_octet_count) {   
    last_update = new Date();

    if (header.getPacketType() != 200) return;

    sender_info = new RtcpReportSendInfo();

    sender_info.ntp_timestamp_most   = ntp_timestamp_most;
    sender_info.ntp_timestamp_least  = ntp_timestamp_least;
    sender_info.rtp_timestamp        = rtp_timestamp;
    sender_info.sender_packet_count  = sender_packet_count;
    sender_info.sender_octet_count   = sender_octet_count;
  }

  /************************************************************************/
  // Set and get Session Description Info fields
  /************************************************************************/

  public void setSdesInfo(int fieldId, String value) {
    last_update = new Date();

    switch(fieldId) {
    case 1: // Cname
      sdesCname=value;
      break;
    case 2: // Name
      sdesName=value;
      break;
    case 3: // Email
      sdesEmail=value;
      break;
    case 4: // Phone
      sdesPhone=value;
      break;
    case 5: // Location
      sdesLoc=value;
      break;
    case 6: // Tool
      sdesTool=value;
      break;
    case 7: // Note
      sdesNote=value;
    case 8: // Priv - don't use
      break;
    default:
      System.out.println("RtcpLogger: discarding SDES field "
                + fieldId + "/value <" + value + ">");
      break;
    }
  }
  public String getSdesCname() { return(sdesCname); }
  public String getSdesName()  { return(sdesName);  }
  public String getSdesEmail() { return(sdesEmail); }
  public String getSdesPhone() { return(sdesPhone); }
  public String getSdesLoc()   { return(sdesLoc);   }
  public String getSdesNote()  { return(sdesNote);  }
  public String getSdesTool()  { return(sdesTool);  }

  public RtcpReportBlock findReportBlock(long ssrc) {
    if (getReportCount() == 0)
      return(null);
    
    for (int x=0; x < getReportCount(); x++) {
      if (report_blocks[x].source_ssrc == ssrc) 
        return(report_blocks[x]);
    }

    return(null);
  }

  public int getStatus(long ssrc) { 
    RtcpReportBlock rrb = findReportBlock(ssrc);
    if (rrb == null)
      return(UNKNOWN);

    if (rrb.getLossPercentage() > 9)
      return(HIGH_LOSS);
    else if (rrb.getLossPercentage() > 1)
      return(MEDIUM_LOSS);
    else
      return(NO_LOSS);
  }
}
