/****************************************************************************
* Tree.java
*
* PURPOSE
* Hold data on an entire MHealth tree.
*
* MHEALTH
* Written by David Makofske and Kevin Almeroth
* University of California at Santa Barbara
* Department of Computer Science
* Neworking and Multimedia Systems Laboratory
* http://imj.ucsb.edu/mcontrol
*
* COPYRIGHT
* Copyright 1998, 1999
* University of California at Santa Barbara
* Networking and Multimedia Systems Lab
* Santa Barbara, CA  93106
* ALL RIGHTS RESERVED
*
* This material may be modified, copied and redistributed, both within
* the recipient's organization and externally, subject to the following
* restrictions:  (a)  The recipient may not derive income from the
* University of California at Santa Barbara (herein "UCSB") information
* itself;  (b)  In any material based on this information, the recipient
* agrees to acknowledge UCSB; (c)  Any copies made of this material must be
* accompanied by the following copyright notice: "Copyright 1998, 1999
* University of California at Santa Barbara. Santa Barbara, CA 93106. All
* Rights Reserved."; and (d)  The recipient agrees to obey all U.S.
* Government restrictions governing redistribution or export of such
* information. These restrictions may apply to redistribution within an
* international organization. UCSB makes no warranties or representations,
* either expressed or implied, with respect to the results containied
* herein, its quality, merchantability, performance or fitness for a
* particular purpose. In no event shall UCSB or its developers, directors,
* officers, employees or affiliates be liable for direct, incidental,
* indirect, special or consequential damages (including damages or loss
* of business profits, business interruption, loss of business information
* and the like) resulting from any defect in this material or its
* documentation or arising out of the use or inability to use this material
* or accompanying documentation even if UCSB, an authorized representative
* or a UCSB affiliate has been advised of the possibility of such damage.
* UCSB makes no representation or warranty regarding the results obtainable
* through use of this material.  No oral or written information or advice
* given by UCSB, its dealers, distributors, agents, affiliates, developers,
* directors, officers or employees shall create a warranty or in any way
* increase the scope of this warranty.
****************************************************************************/

import java.util.*;
import java.io.*;

public class Tree {

  private Hashtable placedReceivers; 	// hashtable of placed receivers w/paths
  private Hashtable tree; 	  	// hashtable of all nodes in tree

  private TreeNode root=null; 		// points to root of tree
  private Vector listener=null;

  private RtcpLogger rtcplogger = null;

  public Tree() {
    root = null;
    placedReceivers = new Hashtable();
    tree = new Hashtable();
    listener = new Vector();
  }

  public void setRoot(TreeNode tn) {
    tn.level = 0;
    root = tn;
    tree.put(root.IP, root);
  }

  public TreeNode getRoot() {
    return root;
  }

  public void addPath(ReceiverPath receiverPath, 
                      ReceiverNode currMtraceReceiverNode) {
    System.out.println("Tree: addPath(" + receiverPath.getReceiverIP() + ")");

    // First check if this receiver is already placed
    ReceiverPath oldPath = (ReceiverPath)placedReceivers.get(
                                                 receiverPath.getReceiverIP());

    if (oldPath == null) {
       // this is a newly placed receiver
       placedReceivers.put(receiverPath.getReceiverIP(), receiverPath);
       addNewPath(receiverPath, currMtraceReceiverNode);
    } else {
       // this is an update of an existing path

       if (oldPath.isSamePath(receiverPath) == true) {
         oldPath.update(receiverPath);
       } else {
         System.out.println("*** PATH CHANGED!!!!!!!!!!!!!");
         removePath(oldPath);
         addNewPath(receiverPath, currMtraceReceiverNode);
         oldPath.update(receiverPath);
       }
    }
  }

  public void removePath(ReceiverPath oldPath) {
    System.out.println("Tree: removePath( " + oldPath.getReceiverIP() + ")");
    TreeNode tn;
    TreeNode holdTN=null;

    for (int x=0; x < oldPath.indexLastHop+1; x++) {
      tn = (TreeNode)tree.get(oldPath.mtraceHop[x].getIP());
      if (tn == null) continue;
 
      if (holdTN != null)
        tn.removeChild(holdTN);

      if (tn.hasChildren()) break;

      holdTN = tn;
      tree.remove(tn.IP);
      
    }
  }

  public void addNewPath(ReceiverPath newPath, 
                         ReceiverNode currMtraceReceiverNode) {
    System.out.println("Tree: addNewPath( " + newPath.getReceiverIP() + ")");

    TreeNode tn=null;
    TreeNode parentTN=null;

    // sanity check
    if (! root.IP.equals(newPath.mtraceHop[newPath.indexLastHop].getIP())) {
      System.out.println("Tree: Path has incorrect root! Actual root " + 
                         root.IP + ", path root " + 
                         newPath.mtraceHop[newPath.indexLastHop].getIP());
      return;
    }
    
    for (int x=newPath.indexLastHop; x >= 0; x--) {
       tn = (TreeNode)tree.get(newPath.mtraceHop[x].getIP());
       if (tn == null) {
         tn = new TreeNode(newPath.mtraceHop[x].getName(),
                           newPath.mtraceHop[x].getIP());
       }
       tn.mtraceStats = new MtraceStats(newPath.getReceiverIP(),
                                        newPath.mtraceStart,
                                        newPath.mtraceFinish);
       if (parentTN != null) {
         tn.addParent(parentTN); 
         parentTN.addChild(tn);
       }
       tree.put(tn.IP, tn);
       parentTN = tn;
    }
    if (newPath.mtraceHop[0].getIP().equals(newPath.getReceiverIP())){
      tn.setRtcpReport(currMtraceReceiverNode.getRtcpReport());
    }
  }

  public void setRtcpLogger(RtcpLogger rtcplogger) {
    this.rtcplogger = rtcplogger;
  }

  public TreeNode findNode(String IP) {
    return((TreeNode)tree.get(IP));
  }

  public TreeNode findInside(TreeNode start, int mouseX, int mouseY) {
    
    if (start == null) return(null);

    if (start.isInside(mouseX, mouseY)) {
      return(start);
    } 

    int count = start.childCount();
    if (count == 0) 
      return(null);
    else {
      int x=0;
      TreeNode found = null;
      while ((found==null)&&(x<count)) {
        found = findInside((TreeNode)start.children.elementAt(x),
                           mouseX, mouseY);
        x++;
      }
      return found;
    }
  }

  public void timeoutReceiverInTree(String IP) {
  }

  public boolean removeReceiverFromTree(String IP) {
    ReceiverPath receiverPath = (ReceiverPath)placedReceivers.get(IP);
    if (receiverPath == null) return(false);

    removePath(receiverPath);
    
    return(true);
  }

  public void pruneNode(TreeNode tn) {
    tn.setExpanded(false);
  }

  public void expandNode(TreeNode tn) {
    tn.setExpanded(true);
    TreeNode p;
    for (int x=0; x < tn.parents.size(); x++) {
      p = (TreeNode)tn.parents.elementAt(x);
      p.checkChildExpansion();
    }
  }

  public void expandChildren(TreeNode tn) {
    TreeNode c;
    for (int x=0; x < tn.children.size(); x++) {
      c = (TreeNode)tn.children.elementAt(x);
      if (!c.isExpanded()) expandNode(c);
    }
  }

  public boolean setMessage(String IP, String msg) {
    TreeNode tn = (TreeNode)tree.get(IP);

    if (tn == null) return(false);

    tn.setMessage(msg);
    return(true);
  }

  public void senderChanged(String IP) {
  }

  public void setMtraceLoss(String IP, String packets_lost, 
                            String total_packets) {
    TreeNode tn = (TreeNode)tree.get(IP);
    if (tn == null) return;

    tn.mtraceStats.setMtraceLoss(packets_lost, total_packets);
  }

  public ReceiverNode findReceiverInside(ReceiverNode start, 
                                         int mouseX, int mouseY) {
    if (start == null) return(null);

    if (start.isInside(mouseX, mouseY)) return(start);

    return(findReceiverInside(start.getNextNode(), mouseX, mouseY));
  }

  public void delete() {
    placedReceivers.clear();
    tree.clear();
    root=null;
  }

}
