/****************************************************************************
* TreeNode.java
*
* PURPOSE
* Hold data on a single node of an MHealth Tree.
*
* MHEALTH
* Written by David Makofske and Kevin Almeroth
* University of California at Santa Barbara
* Department of Computer Science
* Neworking and Multimedia Systems Laboratory
* http://imj.ucsb.edu/mcontrol
*
* COPYRIGHT
* Copyright 1998, 1999
* University of California at Santa Barbara
* Networking and Multimedia Systems Lab
* Santa Barbara, CA  93106
* ALL RIGHTS RESERVED
*
* This material may be modified, copied and redistributed, both within
* the recipient's organization and externally, subject to the following
* restrictions:  (a)  The recipient may not derive income from the
* University of California at Santa Barbara (herein "UCSB") information
* itself;  (b)  In any material based on this information, the recipient
* agrees to acknowledge UCSB; (c)  Any copies made of this material must be
* accompanied by the following copyright notice: "Copyright 1998, 1999
* University of California at Santa Barbara. Santa Barbara, CA 93106. All
* Rights Reserved."; and (d)  The recipient agrees to obey all U.S.
* Government restrictions governing redistribution or export of such
* information. These restrictions may apply to redistribution within an
* international organization. UCSB makes no warranties or representations,
* either expressed or implied, with respect to the results containied
* herein, its quality, merchantability, performance or fitness for a
* particular purpose. In no event shall UCSB or its developers, directors,
* officers, employees or affiliates be liable for direct, incidental,
* indirect, special or consequential damages (including damages or loss
* of business profits, business interruption, loss of business information
* and the like) resulting from any defect in this material or its
* documentation or arising out of the use or inability to use this material
* or accompanying documentation even if UCSB, an authorized representative
* or a UCSB affiliate has been advised of the possibility of such damage.
* UCSB makes no representation or warranty regarding the results obtainable
* through use of this material.  No oral or written information or advice
* given by UCSB, its dealers, distributors, agents, affiliates, developers,
* directors, officers or employees shall create a warranty or in any way
* increase the scope of this warranty.
****************************************************************************/

import java.awt.*;
import java.io.*;
import java.util.*;

public class TreeNode {

  public String name;
  public String IP;
  public int level;
  public String pps;

  public long magicNumber=0;

  public MtraceStats mtraceStats=null;

  public static final int NORMAL = 0;
  public static final int WARNING = 1;
  public static final int CONGESTED = 2;
  public static final int NO_DATA = -1;

  public String message = null;

  public Vector children;
  public Vector parents;

  public boolean timeout=false;
  public boolean drawn=false;
  private int x=0;
  private int y=0;
  private Dimension dimensions;

  private boolean Expanded=true;
  private boolean prunedChildren=false;

  public RtcpReport report;

  // some box drawing color constants
  private Color root_box_color        = Color.green;
  private Color receiver_box_color    = Color.green;
  private Color normal_box_color      = Color.white;
  private Color warning_color         = Color.yellow;
  private Color congested_color       = Color.red;
  private Color no_data_color         = Color.pink;
  private Color timeout_box_color     = Color.lightGray;
  private Color box_outline_color     = Color.black;
  private Color box_text_color        = Color.black;
  private Color msg_text_color        = Color.black;
  private Color pruned_children_color = Color.green;
      
  public TreeNode(String name, String IP) {
    if (IP == null) 
       System.out.println("ERROR: TreeNode: null IP!");

    this.name=name;
    this.IP=IP;

    children = new Vector();
    parents = new Vector();
  
  }

  public void addChild(TreeNode tn) {
    children.addElement(tn);
  }

  public void removeChild(TreeNode tn) {
    int index = children.indexOf(tn);
    if (index == -1)
       return;
    
    tn.removeSubtree();
    children.removeElementAt(index);

    System.gc();
  }

  public void removeSubtree() {
    TreeNode temp;

    for (int i=0; i<children.size(); i++) {
      temp = (TreeNode)children.elementAt(i);
      temp.removeSubtree();
    }
    children.removeAllElements();
  }

  public void addParent(TreeNode parent) {
    parents.addElement(parent);
    if ((parent.level+1) >= level)
      raiseLevel(parent.level+1);
  }

  public void raiseLevel(int NewLevel) {
    TreeNode child;

    try {
    if (level < NewLevel) {
      level = NewLevel;
      for (int i=0; i < childCount(); i++) {
        child = (TreeNode)children.elementAt(i);
        child.raiseLevel(NewLevel + 1);
      }
    }
    } catch (Exception e) {
      e.printStackTrace();
      System.exit(-1);
    }
  }

  public boolean hasChildren() {
    return(!children.isEmpty());
  }

  public boolean isLeaf() {
    return(children.isEmpty() && (! isRoot()));
  }

  public boolean isReceiver() {
    return(report != null);
  }

  public boolean isRoot() {
    return(parents.isEmpty());
  }

  public int childCount() {
    return(children.size());
  }

  public boolean hasParents() {
    return(!parents.isEmpty());
  }

  public int parentCount() {
    return(parents.size());
  }

  public boolean isChild(TreeNode child) {
    TreeNode temp;

    if (!hasChildren()) return(false);
    for (int x=0; x < children.size(); x++) {
      temp = (TreeNode)children.elementAt(x);
      if (temp.IP.equals(child.IP))
        return(true);
    }
    return(false);
  }

  public boolean isParent(TreeNode parent) {
    TreeNode temp;

    if (!hasParents()) return(false);
    for (int x=0; x < parents.size(); x++) {
      temp = (TreeNode)parents.elementAt(x);
      if (temp.IP.equals(parent.IP))
        return(true);
    }
    return(false);
  }

  public void setCoordinates(int x, int y, int width, int height) {
    this.x = x;
    this.y = y;
    dimensions = new Dimension(width, height);
  }

  public TreeNode getChild(int i) {
    return((TreeNode)children.elementAt(i));
  }

  public int getX() { return(x); }
  public void setX(int val) { x = val; }

  public int getY() { return(y); }
  public void setY(int val) { y = val; }

  public Dimension getDimensions() { return(dimensions); }
  public void setDimensions(Dimension val) { dimensions = val; }
  public void setDimensions(int width, int height) { 
    dimensions = new Dimension(width,height); 
  }

  public boolean isInside(int mouseX, int mouseY) {
    if ((mouseX > this.x) && (mouseX < (this.x + dimensions.width)) &&
        (mouseY > this.y) && (mouseY < (this.y + dimensions.height)))
      return(true);
    else
      return(false);
  }

  public Dimension getLineFrom() {
    return(new Dimension(x + (dimensions.width/2), y + dimensions.height));
  }

  public Dimension getLineTo() {
    return(new Dimension(x + (dimensions.width/2), y));
  }

  public boolean isExpanded() { return(Expanded); }
  public void setExpanded(boolean Val) { 
    Expanded = Val; 
    if (Val == false) {
      TreeNode tn;
      for (int x=0; x < parents.size(); x++) {
        tn = (TreeNode) parents.elementAt(x);
        tn.setHasPrunedChildren(true);
      }
    }
  }

  public boolean hasPrunedChildren() { return(prunedChildren); }
  public void setHasPrunedChildren(boolean Val) { prunedChildren = Val; }

  public boolean allChildrenPruned() { 
    TreeNode tn;
    for (int x=0; x < children.size(); x++) {
      tn = (TreeNode) children.elementAt(x);
      if (tn.isExpanded())
        return(false);
    }
    return(true); 
  }

  public boolean getTimeout() { return(timeout); }
  public void setTimeout(boolean Val) { timeout = Val; }

  public void setMessage(String message) { this.message = message; }
  public String getMessage() { return(message); }

  public void setRtcpReport(RtcpReport report) { 
    //Debug.out(debugLevel, Debug.INFO, "TreeNode: setRtcpReport()");
    this.report = report; 
  }

  public RtcpReport getRtcpReport() { 
    return(report);
  }

  public void checkChildExpansion() {
    TreeNode tn;
  
    for (int x = 0; x < children.size(); x++) {
      tn = (TreeNode)children.elementAt(x);
      if (!tn.isExpanded()) { 
        setHasPrunedChildren(true);
        return;
      }
    }
    setHasPrunedChildren(false);
  }

  public void debugDump() {
    TreeNode tn;
    System.out.println("dumpDebug()"); 
    System.out.println(IP + " / " + name); 
    System.out.println("child count = " + children.size()); 
    for (int x = 0; x < children.size(); x++) {
      tn = (TreeNode)children.elementAt(x);
      System.out.println("child (" + x + ") is " + tn.IP + " / " + tn.name);
    }
    System.out.println("parent count = " + parents.size()); 
    for (int x = 0; x < parents.size(); x++) {
      tn = (TreeNode)parents.elementAt(x);
      System.out.println("parent (" + x + ") is " + tn.IP + " / " + tn.name);
    }
  }

  public void draw(Graphics g) {
    FontMetrics fm;
    int font_width;
    int loss;
    int status;

    if (mtraceStats == null) {
      status = NORMAL;
    } else {
      loss = mtraceStats.getMtraceLossPercent();
      if (loss > 9) 
        status = CONGESTED;
      else if (loss > 2) 
        status = WARNING;
      else
        status = NORMAL;
      if (report != null) {
        if (report.getReportCount() == 0) 
          status = NO_DATA;
        else if (report.report_blocks[0].fraction_lost > 20)
          status = CONGESTED;
        else if (report.report_blocks[0].fraction_lost > 5)
          status = WARNING;
        else
          status = NORMAL;
      }
    }

    if (getTimeout())
      g.setColor(timeout_box_color);
    else if (isRoot())
      g.setColor(root_box_color);
//    else if (isLeaf())
//      g.setColor(receiver_box_color);
    else  {
      switch (status) {
      case NORMAL:
        if (isReceiver())
          g.setColor(receiver_box_color);
        else
          g.setColor(normal_box_color);
        break;
      case WARNING:
        g.setColor(warning_color);
        break;
      case CONGESTED:
        g.setColor(congested_color);
        break;
      case NO_DATA:
        g.setColor(no_data_color);
        break;
      }
    }

    g.fillRect(x, y, dimensions.width, dimensions.height);

//    System.out.println("draw me " + name + " at " + x + "," + y);
    // draw box outline
    g.setColor(box_outline_color);
    g.drawRect(x, y, dimensions.width, dimensions.height);

    if (hasPrunedChildren()) {
      g.setColor(pruned_children_color);
      g.fillRect(x, y + dimensions.height-3, dimensions.width, 3);
    }

    g.setColor(box_text_color);
    fm = g.getFontMetrics(g.getFont());
    font_width = fm.stringWidth(name);
    if (fm.stringWidth(name) > dimensions.width) {
      // name doesn't fit in box, let's try IP address
      font_width = fm.stringWidth(IP);
      g.drawString (IP, x + ((dimensions.width-font_width)/2), y+15);
    } else {
      font_width = fm.stringWidth(name);
      g.drawString (name, x + ((dimensions.width-font_width)/2), y+15);
    }

    if ((mtraceStats != null) && 
//        (! mtraceStats.getMtraceLostPercentStr().equals("?")) &&
        (! isLeaf())) {
      String strLoss = mtraceStats.getMtraceLoss();
      font_width = fm.stringWidth(strLoss);
      switch (status) {
      case NORMAL:
        g.setColor(box_text_color);
        break;
      case WARNING:
//        g.setColor(warning_color);
//        g.fillRect(x + (dimensions.width/2), y+21, font_width + 4, 12);
        g.setColor(box_text_color);
        break;
      case CONGESTED:
//        g.setColor(congested_color);
//        g.fillRect(x + (dimensions.width/2), y+21, font_width + 4, 12);
        g.setColor(box_text_color);
        break;
      }
      g.drawString (strLoss, x + (dimensions.width/2)+ 2, y+30);
//      font_width = fm.stringWidth(strLoss);
//      g.drawString (strLoss, x + ((dimensions.width-font_width)/2), y+30);
    }

    if (message != null) {
      g.setColor(msg_text_color);
      font_width = fm.stringWidth(message);
      g.drawString (message, x + ((dimensions.width-font_width)/2), y+30);
    }
  }
}
